#!/usr/bin/env pythonbutton_color
# -*- coding: utf-8 -*-

"""  

Higher level ui components / widgets like Edit_Panel, Diagram 

"""

import os

from copy               import copy
from typing             import override, Callable

from PyQt6.QtCore       import Qt, QEvent, QPoint, QRectF
from PyQt6.QtCore       import QSize, QMargins, QTimer, QPropertyAnimation, QAbstractAnimation
from PyQt6.QtWidgets    import (QLayout, QGridLayout, QVBoxLayout, QHBoxLayout, QSizePolicy,
                                QWidget, QDialog, QDialogButtonBox, QLabel, QMessageBox, QFrame,
                                QGraphicsOpacityEffect, QTabWidget)
from PyQt6.QtGui        import QGuiApplication, QScreen, QColor, QPalette, QPainterPath, QRegion, QTransform
from PyQt6              import sip

from .widgets           import set_background, style
from .widgets           import Widget, Label, CheckBox, size, Icon

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


#------------------------------------------------------------------------------
# Utils for QMainWindow and QDialog  
#------------------------------------------------------------------------------

class Win_Util: 
    """ 
    Utility functions for window handling 
    """

    @staticmethod
    def set_initialWindowSize (qwindow : QWidget,
                               size : tuple | None = None,
                               size_frac : tuple | None = None,
                               pos : tuple | None = None,
                               pos_frac: tuple | None = None,
                               geometry : tuple | None = None,
                               maximize : bool = False):
        """
        Set size and position of Qt window in fraction of screensize or absolute
        """

        # geometry argument has priority 

        if geometry: 
            qwindow.setGeometry (*geometry)
            if maximize:
                qwindow.showMaximized()
            return
        else:  
            x, y, width, height = None, None, None, None
 
        # set size 

        if size_frac: 

            screen : QScreen = QGuiApplication.primaryScreen()
            screenGeometry = screen.geometry()
 
            width_frac, height_frac  = size_frac

            if width_frac:   width  = screenGeometry.width()  * width_frac
            if height_frac:  height = screenGeometry.height() * height_frac

        if size:
            width, height = size

        width  = int (width)  if width  is not None else 1000
        height = int (height) if height is not None else  700
        
        qwindow.resize (QSize(width, height))

        if maximize: 
            qwindow.showMaximized()

        # set position 

        if pos: 
            x, y = pos

        if pos_frac: 

            screen : QScreen = QGuiApplication.primaryScreen()
            screenGeometry = screen.geometry()
 
            x_frac = pos_frac[0]
            y_frac = pos_frac[1]
            if x_frac: x = screenGeometry.width()  * x_frac
            if y_frac: y = screenGeometry.height() * y_frac

        x = int (x) if x  is not None else 200
        y = int (y) if y is not None else  200
        
        qwindow.move (x, y)


#------------------------------------------------------------------------------

class Disabled_Overlay(QWidget):
    """ Create a semi-transparent overlay to indicate disabled state """

    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.setGeometry(parent.rect())
        set_background (self, color=QColor("#202020"), alpha=0.7)
        self.raise_()                               # Ensure it's on top
        self.show()
    
    def resizeEvent(self, event):
        # Keep overlay covering the entire parent window
        self.resize(self.parent().size())



#------------------------------------------------------------------------------
# Panels - QWidgets like a field group within a context 
#------------------------------------------------------------------------------


class Panel_Abstract (QWidget):
    """ 
    Superclass for other types of panels like Edit or Container
        - handle size of widget  
        - having title / header name 
        - has dataObject via getter (callable) 
    """

    name = "Panel"                          # will be title 

    _width  = None
    _height = None 

    _n_doubleClick = 0                      # no of double clicks - common counter for all panels
    _doubleClick_hint : str|None = None     # hint for double click at the end of box layout for all panels


    # ------------------------------------------------------

    @staticmethod
    def widgets_of_layout  (layout: QLayout) -> list[Widget]:
        """ list of Widgets defined in layout"""

        # iterate over layout - not self._panel.findChildren (Widget) as widgets are "deleted later"
        #       and can still exist although they are no more in the layout 

        widgets = []
        if isinstance (layout,QLayout): 
            for i in range(layout.count()):
                widget = layout.itemAt(i).widget()
                if isinstance (widget, Widget):                
                    widgets.append(widget)
                elif isinstance (widget,QWidget):                       # could be helper QWidget with seperate layout
                    widgets.extend (Panel_Abstract.widgets_of_layout (widget.layout()))
        return widgets


    @staticmethod
    def panels_of_layout  (layout: QLayout) -> list['Edit_Panel']:
        """ list of first level Edit_Panels defined in layout"""

        # iterate over layout - not self._panel.findChildren (Widget) as widgets are "deleted later"
        #       and can still exist although they are no more in the layout 

        panels = []
        if isinstance (layout,QLayout): 
            for i in range(layout.count()):
                widget = layout.itemAt(i).widget()
                if isinstance (widget, Edit_Panel):                
                    panels.append(widget)
        return panels


    # ------------------------------------------------------

    def __init__(self,  
                 app = None,
                 getter : Callable|None = None,                 # data object or callable returning it
                 width : int|None = None, 
                 height : int|None = None, 
                 hide : Callable|bool|None = None,              # either callable or bool
                 doubleClick: Callable|None = None,             # callable when doubleClick event occurs
                 hint : str|None = None,                        # hint for double click at the end of box layout
                 title=None, **kwargs):                         # title of panel (default self.name)

        super().__init__( **kwargs)

        self._app = app
        self._getter = getter

        if width is not None: 
            self._width = width
        if height is not None: 
            self._height = height

        # handle visibility  
        self._shouldBe_visible = True                               # default visibility of self 
        self._hidden_fn = None                                      # altern. callable to set visibility
        if isinstance(hide, bool):
            self._shouldBe_visible = not hide
        elif callable (hide):
            self._hidden_fn = hide

        # set width and height 
        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        if title is not None: 
            self.name = title 

        # handle double click event
        self._doubleClick : callable|None = doubleClick
        self._doubleClick_hint_widget : Label | None = None
        Panel_Abstract._doubleClick_hint : str|None = hint          # class variable - common for all instances


    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        name = self.name if self.name else type(self).__name__
        return f"<Panel '{name}'>"


    @override
    def mouseDoubleClickEvent(self, event):
        """ default double click event - call self._doubleClick if defined"""

        if callable(self._doubleClick):
            Panel_Abstract._n_doubleClick += 1                  # count no of double clicks
            if Panel_Abstract._n_doubleClick == 2:              # remove hint when user learned to doubleclick
                Panel_Abstract._doubleClick_hint = None         # class variable - common for all instances
            if self._doubleClick_hint_widget:
                self._doubleClick_hint_widget.refresh()

            self._doubleClick()                                 # callback to parent
        else:
            super().mouseDoubleClickEvent(event)

    @property
    def dataObject (self): 
        # to be overloaded - or implemented with semantic name 
        if callable(self._getter):
            return self._getter()
        else: 
            return self._getter

    @property 
    def shouldBe_visible (self) -> bool:
        """ True if self should be visible 
            - can be overridden to control visibility in subclass """
        
        if callable (self._hidden_fn):
            self._shouldBe_visible = not self._hidden_fn ()
        return self._shouldBe_visible


    def set_visibilty (self, aBool : bool):
        """ 
        set the visibility of self 
            - use this, when instances of Edit_Panel are used (not subclassing)
              to control hide/show
        """

        if self._shouldBe_visible != aBool:
            self._shouldBe_visible = aBool        
            self.setVisible (aBool)     


    def set_doubleClick (self, doubleClick: Callable|None):
        """ set double click callable """
        self._doubleClick = doubleClick


    @property 
    def _isDisabled (self) -> bool:
        """ True if the widgets of self are disabled  - can be overloaded """
        return False

    def set_background_color (self, darker_factor : int | None = None,
                                    color : QColor | int | None  = None,
                                    alpha : float | None = None):
        """ 
        Set background color of a QWidget either by
            - darker_factor > 100  
            - color: QColor or string for new color
            - alpha: transparency 0..1 
        """
        set_background (self, darker_factor=darker_factor, color=color, alpha=alpha)


#-------------------------------------------

class Container_Panel (Panel_Abstract):
    """ 
    Panel as container for other (Edit) panels  
    """

    def __init__(self, *args, 
                 layout : QLayout | None = None,
                 margins  : tuple = (0,0,0,0), 
                 spacing : int = 5, 
                 **kwargs):
        super().__init__(*args, **kwargs)

        if layout:
            layout.setContentsMargins (QMargins(*margins))  
            layout.setSpacing (spacing) 
            self.setLayout (layout)

        # add hint for double click at the end of box layout

        if self._doubleClick_hint is not None:
            self.add_hint (self._doubleClick_hint)

        # initial visibility 

        if not self.shouldBe_visible:         
            self.setVisible (False)                     # setVisible(True) results in a dummy window on startup 


    def calc_min_width (self) -> int:
        """ calculate minimum width of self based on visible panels of self layout """

        layout = self.layout()

        # only QHBoxLayout supported here
        if not isinstance (layout, QHBoxLayout):
            return

        min_width = 0
        ncols = layout.count()
        ncols_filled = 0

        # loop over rows to find maximum row height
        for i in range(ncols):
            col_width = 0
            item = layout.itemAt(i)
            if item:
                qwidget = item.widget()
                if isinstance (qwidget, Panel_Abstract) and not qwidget.isHidden():
                    col_width = max(col_width, qwidget.width()) 
                elif isinstance (qwidget, Widget) and not qwidget.isHidden():
                    col_width = max(col_width, qwidget.minimumWidth()) 
                # print (f"panel {qwidget} width: {qwidget.width()} minWidth: {qwidget.minimumWidth()}" )
            min_width += col_width
            if col_width > 0:
                ncols_filled += 1

        # add horizontal spacings + margins
        margins = layout.contentsMargins()
        min_width = min_width + (ncols_filled-1) * layout.spacing() + margins.left() + margins.right()

        return min_width


    def add_hint (self, hint : str):
        """ add stretchable hint for double click at the end of layout """

        Panel_Abstract._doubleClick_hint = hint         # class variable - common for all instances
        l = self.layout()

        if self._doubleClick_hint_widget is not None:
            self._doubleClick_hint_widget.refresh()

        elif hint is not None and isinstance(l, (QHBoxLayout, QVBoxLayout)):
            w = Label (l, get=lambda: Panel_Abstract._doubleClick_hint, width=(1, None),
                      style=style.HINT, align=Qt.AlignmentFlag.AlignCenter)
            l.setStretch (l.count()-1,2)
            self._doubleClick_hint_widget = w

        self.setToolTip (hint)
              


    @property
    def edit_panels (self) -> list['Edit_Panel']:
        """ list of first level Edit_Panels defined in self"""

        return self.panels_of_layout (self.layout())


    def refresh (self):
        """ refresh all child Panels self"""

        show_now = self.isHidden () and self.shouldBe_visible

        self.setVisible (self.shouldBe_visible)

        if self.shouldBe_visible:

            # first hide the now not visible panels so layout won't be stretched
            for p in self.edit_panels:
                if not p.shouldBe_visible: p.refresh() 

            # now show the now visible panels - reinit layout if needed (with lazy there can be no layout yet)
            for p in self.edit_panels:
                reinit = show_now or (self.shouldBe_visible and p._panel.layout() is None)
                if p.shouldBe_visible: p.refresh (reinit_layout=reinit) 

            # refresh double click hint
            if self._doubleClick_hint_widget:
                self._doubleClick_hint_widget.refresh()
            if self.toolTip() != self._doubleClick_hint:
                self.setToolTip (self._doubleClick_hint)


class Edit_Panel (Panel_Abstract):
    """ 
    Panel with/out a title and an optional on/off switch 
    having a layout area for content  
    """

    _height = (None, None) 

    _main_margins  = (10, 5,10, 5)             
    _head_margins  = ( 0, 0, 0, 5)             
    _panel_margins = (10, 0, 0, 0)           

    def __init__(self, *args, 
                 layout : QLayout | None = None,
                 lazy = False, 
                 switchable : bool = False,                 # if True add on/off checkbox switch to header
                 switched_on : bool | Callable = True,      # initial switch state on/off or callable returning it
                 on_switched : Callable | None = None,      # callback when user switched on/off
                 hide_switched : bool = True,               # if True hide panel area when switched off
                 has_head : bool = True,
                 auto_height : bool = False,                # if True fix min height after layout init
                 main_margins  : tuple = None,             
                 head_margins  : tuple = None,             
                 panel_margins : tuple = None,          
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._switchable = switchable
        self._hide_switched = hide_switched
        self._switched_on = switched_on
        self._on_switched = on_switched
        self._auto_height = auto_height

        self._head  = None
        self._panel = None

        self._main_margins  = main_margins  if isinstance(main_margins, tuple)  else self._main_margins
        self._head_margins  = head_margins  if isinstance(head_margins, tuple)  else self._head_margins   
        self._panel_margins = panel_margins if isinstance(panel_margins, tuple) else self._panel_margins

        # set background color depending light/dark mode

        if Widget.light_mode:
            self.set_background_color (darker_factor = 104)                 # make it darker 
        else: 
            self.set_background_color (darker_factor = 80)                  # make it lighter 

        # title layout - with optional on/off switch 

        if has_head and self.title_text() is not None: 

            self._head = QWidget(self)

            if not lazy: 
                self._set_header_layout ()

 
        # inital content panel content - layout in >init  

        self._panel  = QWidget() 
        self._initial_layout = layout 

        if not lazy: 
            self._set_panel_layout () 

        # main layout with title and panel 

        l_main   = QVBoxLayout()
        if self._head: 
            l_main.addWidget (self._head)
        l_main.addWidget (self._panel, stretch=1)
        l_main.setContentsMargins (QMargins(*self._main_margins))  
        l_main.setSpacing(2)
        self.setLayout (l_main)

        # initial switch state 
        switch_on = self._switched_on () if callable(self._switched_on) else self._switched_on
        self.set_switched_on (switch_on, silent=True)

        # initial enabled/disabled state
        if self._isDisabled: 
            self.refresh_widgets (self._isDisabled) 

        # initial visibility 
        if not self.shouldBe_visible:         
            self.setVisible (False)             # setVisible(True) results in a dummy window on startup 


    def title_text (self) -> str: 
        """ returns text of title - default self.name"""
        # can be overwritten 
        return self.name 

    @override
    def mouseDoubleClickEvent(self, a0):
        """ default double click event - toggle switch if switchable"""

        if self._switchable and not callable(self._doubleClick):
            self.toggle_switched()
        else:
            return super().mouseDoubleClickEvent(a0)

    def _adjust_height (self, switched_on: bool = True):
        """ adjust height of self when switched on/off """

        if self._switchable and self._hide_switched and not switched_on:
            height_new = 35
        elif self._panel.layout() is None:
            height_new = 35                          # minimal height if no layout yet
        else:
            height_new = self._height

        if self.height() != height_new:
            Widget._set_height (self, height_new)


    @property 
    def switched_on (self) -> bool:
        """ True if self is switched on"""
        
        if callable(self._switched_on):
            return self._switched_on()
        else:
            return self._switched_on
    
    def set_switched_on (self, aBool : bool, silent=False):
        """ switch on/off 
            - optional hide main panel 
            - option callback on_switched
        """
        if not callable(self._switched_on):
            self._switched_on = aBool is True 
        
        if self._hide_switched:
            self._panel.setVisible (aBool)

        self._adjust_height (aBool)

        # ensure refresh of panel if made visible again
        if aBool:
            self.refresh_panel (self._isDisabled)

        # set by checkbox - callback to Diagram_Item or parent panel
        if not silent and callable (self._on_switched):              
            self._on_switched (aBool)

        # refresh also header checkbox
        if not silent:
            for w in self.header_widgets:
                w.refresh ()


    def toggle_switched (self):
        """ toggle switched on/off """
        self.set_switched_on (not self.switched_on)

    @property
    def widgets (self) -> list[Widget]:
        """ list of Widgets defined in self panel area"""

        return self.widgets_of_layout (self._panel.layout())


    @property
    def header_widgets (self) -> list[Widget]:
        """ list of widgets defined in self header area"""

        if self._head: 
            return self.widgets_of_layout (self._head.layout())
        else: 
            return []
 

    def refresh(self, reinit_layout=False):
        """ refreshes all Widgets on self """

        hide = not self.shouldBe_visible and     self.isVisible()
        show =     self.shouldBe_visible and not self.isVisible()

        # hide / show self 

        if hide or show: 
            self.setVisible (self.shouldBe_visible)
            logger.debug (f"{self} - setVisible ({self.shouldBe_visible})")

        if not self.shouldBe_visible: 
            return                                      # no more refresh if hidden 

        # header 
        if self._head:
            self._set_header_layout()                   # ensure header layout
            self.refresh_header ()

        # panel 

        should_hide_panel = self._switchable and not self.switched_on and self._hide_switched

        if should_hide_panel:
            self._panel.setVisible (False)
        else:
            self._panel.setVisible (True)

            # reinit layout if forced or show again after hide 
            #   - but only if layout is from subclass - not from init argument
            if (show or reinit_layout) and not self._initial_layout:                   # show again or forced reinit
                self._set_panel_layout ()
                logger.debug (f"{self} - refresh - reinit_layout ")

            # normal refresh 
            else:
                self.refresh_panel (self._isDisabled, reinit_layout=reinit_layout)
                logger.debug (f"{self} - refresh ")

        self._adjust_height (self.switched_on)



    def refresh_header (self):
        """ refresh all header widgets of self """

        for w in self.header_widgets:
            w.refresh ()


    def refresh_panel (self, disable : bool, reinit_layout=False):
        """ refresh all panel widgets including child panels of self """

        for w in self.widgets:
            w.refresh (disable=disable)

        # refresh child edit panels like Polar_Panel
        for p in self.panels_of_layout (self._panel.layout()):
            p.refresh (reinit_layout=reinit_layout)


    def refresh_widgets (self, disable : bool, reinit_layout=False):
        """ refresh all widgets of self including  sub panels """

        self.refresh_panel (disable, reinit_layout=reinit_layout)

        # refresh also header 
        self.refresh_header()


    def _set_header_layout(self):
        """ set layout of self._head 
            - with title text either checkbox or label
            - a subclass may add widgets to header layout in _add_to_header_layout
        """

        if self._head.layout(): return       # already set

        l_head = QHBoxLayout()
        l_head.setContentsMargins (*self._head_margins)

        if self._switchable:
            CheckBox (l_head, fontSize=size.HEADER, text=self.title_text(),
                    get=lambda: self.switched_on, set=self.set_switched_on,
                    toolTip='Show or hide -<br>Alternatively, you can double click on the panel')
        else: 
            Label (l_head, fontSize=size.HEADER, get=self.title_text)
        l_head.setStretch (0,1)

        self._add_to_header_layout (l_head)     # optional individual widgets

        # assign layout to head widget
        self._head.setLayout (l_head)



    def _set_panel_layout (self):
        """ 
        Set layout of self._panel   
            - typically defined in subclass in _init_layout
            - or as init argument 'layout'"""

        # remove and rebuild only in case of _init_layout done in a subclass

        if isinstance(self._initial_layout, QLayout):

            layout = self._initial_layout

        else:
            if self._panel.layout() is not None:
                self._clear_existing_panel_layout ()

            if self.shouldBe_visible:
                layout = self._init_layout()        # subclass will create layout 
            else: 
                # if the panel shouldn't be visible repalce the normal panel layout
                # with a dummy, so get/set of the normal panel widgets won't get trouble 
                layout = QVBoxLayout()               
                wdt = QLabel ("This shouldn't be visible")
                layout.addWidget (wdt)

        if layout:
            layout.setContentsMargins (QMargins(*self._panel_margins))   # inset left 

            # set default spacings 
            layout.setSpacing (2)
            if isinstance (layout, QGridLayout):
                layout.setVerticalSpacing(4)

            self._panel.setLayout (layout)

            # if height of panel to height of layout with all the widgets
            if self._auto_height:
                self._set_auto_height_panel ()


    def _clear_existing_panel_layout(self):
        """ removes all items from the existing panel layout"""
        layout = self._panel.layout()
        if layout is not None: 
            while layout.count():
                child = layout.takeAt(0)
                childWidget = child.widget()
                if childWidget:                
                    childWidget.deleteLater() # sip.delete (childWidget)
                sip.delete (child)
                # if childWidget:
                #     childWidget.disconnect()
                #     childWidget.setParent(None)
                #     sip.delete (childWidget)
                #     # childWidget.deleteLater()                             # will create ghost widget due to async event loop
            logger.debug (f"{self} - clear layout ")

            # finally remove self 
            sip.delete (self._panel.layout())


    def _set_auto_height_panel (self):
        """ fix minimum height of self._panel to nrows of widgets """

        layout = self._panel.layout()

        # only GridLayout supported here
        if not isinstance (layout, QGridLayout):
            return

        min_height = 0
        nrows = layout.rowCount()
        ncols = layout.columnCount()
        nrows_filled = 0

        # loop over rows to find maximum row height
        for row in range(nrows):
            row_height = 0
            for col in range(ncols):
                item = layout.itemAtPosition(row, col)
                if item:
                    widget = item.widget()
                    if isinstance (widget, Widget):
                        if not widget._hidden:
                            row_height = max(row_height, widget.height()) 
                    elif isinstance(widget, QWidget) and widget.isVisible():
                        row_height = max(row_height, widget.height())
            min_height += row_height
            if row_height > 0:
                nrows_filled += 1

        # add vertical spacings and margins
        min_height = min_height + (nrows_filled-1) * layout.verticalSpacing() + self._panel_margins[1] + self._panel_margins[3]

        if min_height > 0:
            logger.debug (f"{self} - fix min height of panel to {min_height} ")

            self._panel.setMinimumHeight (min_height)


    def _init_layout(self) -> QLayout:
        """ init and return main layout"""

        # to be implemented by sub class
        pass


    def _add_to_header_layout(self, l_head : QHBoxLayout):
        """ add Widgets to header layout"""

        # to be implemented by sub class
        pass



# ------------ MessageBox  -----------------------------------


class MessageBox (QMessageBox):
    """ 
    Subclass of QMessagebox 
        - new default icons 
        - more width and height 
    """

    _min_width  = 250
    _min_height = 80 

    def __init__(self, parent: object, 
                 title : str, 
                 text : str, 
                 icon: Icon, 
                 min_width=None,                            # width of text widget 
                 min_height=None):                          # height of text widget 
        super().__init__(parent)

        # set properties 

        self.setWindowTitle (title)
        self.setText (text)

        # set icon 

        if isinstance (icon, Icon):
            pixmap = icon.pixmap((QSize(32, 32)))
            self.setIconPixmap (pixmap)

        # set width and height 
        #   size of QMessageBox must be set via layout - which is a bit hacky 

        layout : QGridLayout = self.layout()
        if isinstance (layout, QGridLayout):

            cols = layout.columnCount()
            if cols > 1:
                min_width = min_width if min_width is not None else self._min_width
                # set minimum width of last column (which should be text) 
                layout.setColumnMinimumWidth (cols-1,min_width)
                layout.setColumnStretch (cols-1,3)
                # set minimum width of first column (which should be icon) 
                layout.setColumnMinimumWidth (0,60)
                item = layout.itemAtPosition (0,1)
                item.setAlignment (Qt.AlignmentFlag.AlignVCenter )

                # increase right margin 
                left, top, right, bottom = layout.getContentsMargins()
                layout.setContentsMargins (left, top, right+20, bottom)

            rows = layout.columnCount()
            if rows > 1:
                min_height = min_height if min_height is not None else self._min_height
                # set minimum widthof last column (which should be text) 
                layout.setRowMinimumHeight (0,min_height)
                # set center alignment of icon 
                item = layout.itemAtPosition (0,0)
                item.setAlignment (Qt.AlignmentFlag.AlignCenter )

        if os.name == 'posix':
            # strange posix - wayland bug 
            #   With first call of MessageBox the box is positioned at 0,0 -> manually move to the center of parent 
            point = parent.geometry().center()
            x = point.x() - 200
            y = point.y() - 100
            self.move(x,y)
 

    @staticmethod
    def success (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ success message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.SUCCESS), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def info (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ info message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.INFO), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def error (parent: object, title : str, text : str, min_width=None, min_height=None):
        """ critical message with Ok button"""

        msg = MessageBox (parent, title, text, Icon (Icon.ERROR), min_width=min_width, min_height= min_height)
        msg.exec()


    @staticmethod
    def confirm (parent: object, title : str, text : str, min_width=None):
        """ confirmation with Ok and Cancel"""

        msg = MessageBox (parent, title, text, Icon (Icon.INFO), min_width=min_width)

        msg.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
        msg.setDefaultButton  (QMessageBox.StandardButton.Ok)

        return msg.exec()


    @staticmethod
    def warning (parent: object, title : str, text : str, min_width=None):
        """ warning with Ok and Cancel"""

        msg = MessageBox (parent, title, text, Icon (Icon.WARNING), min_width=min_width)

        msg.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
        msg.setDefaultButton  (QMessageBox.StandardButton.Ok)

        return msg.exec()


    @staticmethod
    def save (parent: object, title : str, text : str, min_width=None, buttons=None):
        """ ask to save or discard - returns QMessageBox.StandardButton"""

        msg = MessageBox (parent, title, text, Icon (Icon.WARNING), min_width=min_width)

        if buttons is None: 
            buttons = QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel 
        msg.setStandardButtons(buttons)
        msg.setDefaultButton  (QMessageBox.StandardButton.Save)

        return msg.exec()



# ------------ Toaster  -----------------------------------


class Toaster (QFrame):

    """ 
    Show a little notification toaster 

    based on https://stackoverflow.com/questions/59251823/is-there-an-equivalent-of-toastr-for-pyqt
    """

    def __init__(self, *args, **kwargs):

        super (Toaster, self).__init__(*args, **kwargs)
        QHBoxLayout(self)

        self.label = None

        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)

        #                 border: 1px solid blue;
        # background: palette(window);

        # self.setStyleSheet('''
        #     QToaster {
        #         border-radius: 2px; 
        #     }
        # ''')
        # alternatively:

        self.setAutoFillBackground(True)
        # set_background (self, color="red")
        # self.setFrameShape(self.Box)

        self.timer = QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.hide)

        if self.parent():
            self.opacityEffect = QGraphicsOpacityEffect(opacity=0)
            self.setGraphicsEffect (self.opacityEffect)
            self.opacityAni = QPropertyAnimation (self.opacityEffect, b'opacity')
            # we have a parent, install an eventFilter so that when it's resized
            # the notification will be correctly moved to the right corner
            self.parent().installEventFilter(self)
        else:
            # there's no parent, use the window opacity property, assuming that
            # the window manager supports it; if it doesn't, this won'd do
            # anything (besides making the hiding a bit longer by half a second)
            self.opacityAni = QPropertyAnimation(self, b'windowOpacity')
        self.opacityAni.setStartValue(0.)
        self.opacityAni.setEndValue(1.)
        self.opacityAni.setDuration(100)
        self.opacityAni.finished.connect(self.checkClosed)

        self.corner = Qt.Corner.TopLeftCorner
        self.margin = QMargins(10, 10, 10, 10)


    def checkClosed(self):
        # if we have been fading out, we're closing the notification
        if self.opacityAni.direction() == QAbstractAnimation.Direction.Backward:
            self.close()

    def restore(self):
        # this is a "helper function", that can be called from mouseEnterEvent
        # and when the parent widget is resized. We will not close the
        # notification if the mouse is in or the parent is resized
        self.timer.stop()
        # also, stop the animation if it's fading out...
        self.opacityAni.stop()
        # ...and restore the opacity
        if self.parent():
            self.opacityEffect.setOpacity(1)
        else:
            self.setWindowOpacity(1)

    def hide(self):
        # start hiding
        self.opacityAni.setDirection(QAbstractAnimation.Direction.Backward)
        self.opacityAni.setDuration(300)
        self.opacityAni.start()

    def eventFilter(self, source, event : QEvent):
        if source == self.parent() and event.type() == QEvent.Type.Resize:

            self.opacityAni.stop()

            parentRect : QRectF = self.parent().rect()
            geo = self.geometry()

            corner = self.corner
            margin = self.margin

            if corner == Qt.Corner.TopLeftCorner:
                geo.moveTopLeft(parentRect.topLeft() + QPoint(margin.left(), margin.top()))
            elif corner == Qt.Corner.TopRightCorner:
                geo.moveTopRight(parentRect.topRight() + QPoint(-margin.right(), margin.top()))
            elif corner == Qt.Corner.BottomRightCorner:
                geo.moveBottomRight(parentRect.bottomRight() + QPoint(-margin.right(), -margin.bottom()))
            else:
                geo.moveBottomLeft(parentRect.bottomLeft() + QPoint(margin.left(), -margin.bottom()))

            self.setGeometry(geo)
            self.restore()
            self.timer.start()
        return super(Toaster, self).eventFilter(source, event)

    # def enterEvent(self, event):
    #     deactivated - when the mouse is by accident over the notification, 
    #           the toaster would wouldn't close
    #     self.restore()
    #     super().enterEvent(event)


    def leaveEvent(self, event):
        self.timer.start()

    def closeEvent(self, event):
        # we don't need the notification anymore, delete it!
        self.deleteLater()

    def resizeEvent(self, event):
        super(Toaster, self).resizeEvent(event)
        # if you don't set a stylesheet, you don't need any of the following!
        if not self.parent():
            # there's no parent, so we need to update the mask
            path = QPainterPath()
            path.addRoundedRect(QRectF (self.rect()).translated(-.5, -.5), 4, 4)
            self.setMask(QRegion(path.toFillPolygon(QTransform()).toPolygon()))
        else:
            self.clearMask()

    @staticmethod
    def showMessage(parent, message, 
                    corner = Qt.Corner.BottomLeftCorner, 
                    margin =          QMargins(10,10,10,10), 
                    contentsMargins = QMargins(30, 7,30, 7),
                    toast_style = style.HINT,
                    duration = 1500,
                    alpha :int = 255,
                    parentWindow = False):
        """
        show a toaster for a while

        Args:
            parent : parent Toasters position is relative 
            message: message to show 
            corner : position in parents window 
            margin : offset of position
            style : color style 
            duration: duration of message 
            parentWindow : take parent window of parent
        """

        if parent and parentWindow:
            parent = parent.window()

        if isinstance (margin, int):
            margin = QMargins(margin, margin, margin, margin)

        if toast_style is None: 
            toast_style = style.HINT

        self = Toaster(parent)
        parentRect : QRectF = parent.rect()

        self.timer.setInterval(duration)

        self.label = QLabel(message)
        self.layout().addWidget(self.label)
        self.layout().setContentsMargins (contentsMargins)

        # set background color style 

        if toast_style in [style.WARNING, style.ERROR, style.COMMENT, style.GOOD, style.HINT]:

            palette : QPalette = self.palette()
            index = Widget.LIGHT_INDEX if Widget.light_mode else Widget.DARK_INDEX
            color = QColor (toast_style.value[index])

            # adapt color 

            h,s,l,a = color.getHsl()
            a = alpha if isinstance(alpha,int) else a
            if Widget.light_mode:
                s = int(s*0.8)                                  # less saturation (more gray) 
                l = int(l*1.4)                                  # lighter 
            else:
                s = int(s*0.8)                                  # less saturation (more gray) 
                l = int(l*0.5)                                  # darker 
            color.setHsl (h, s, l, a)

            palette.setColor(QPalette.ColorRole.Window, color)
            self.setPalette(palette)

        # start hide timer 

        self.timer.start()

        # raise the widget and adjust its size to the minimum

        self.raise_()
        self.adjustSize()

        self.corner = corner
        self.margin = margin

        geo = self.geometry()

        # now the widget should have the correct size hints, let's move it to the right place
        
        if corner == Qt.Corner.TopLeftCorner:
            geo.moveTopLeft(parentRect.topLeft() + QPoint(margin.left(), margin.top()))
        elif corner == Qt.Corner.TopRightCorner:
            geo.moveTopRight(parentRect.topRight() + QPoint(-margin.right(), margin.top()))
        elif corner == Qt.Corner.BottomRightCorner:
            geo.moveBottomRight(parentRect.bottomRight() + QPoint(-margin.right(), -margin.bottom()))
        else:
            geo.moveBottomLeft(parentRect.bottomLeft() + QPoint(margin.left(), -margin.bottom()))

        self.setGeometry(geo)
        self.show()
        self.opacityAni.start()



# ------------ Dialog  -----------------------------------

class Dialog (QDialog):
    """
    Abstract super class for modal top windows with action buttons at bottom.
    Extends QDialog with a dataObject (via 'getter') and common background color.

    The content of the dialog is defined in a subclass via _init_layout() which returns the
    layout for the dialog.
    """

    name = "Dialog"             # will be title 

    _width  = None
    _height = None 


    def __init__(self,  
                 parent : QWidget = None,
                #  flags : Qt.WindowType = None, 
                 getter = None, 
                 width  : int | None =None,             # optional width of dialog
                 height : int | None =None,             # optional height of dialog
                 dialogPos : tuple = (0.4,0.5),         # reference point for positioning - (0,0) is upper, left corner
                 parentPos : tuple = (0.3,0.6),         # position anchor in parent - (0,0) is upper, left corner
                 dx : int | None = None,                # optional move dialog dx pixels to the right
                 dy : int | None = None,                # optional move dialog dy pixels down
                 title=None,                            # optional title of dialog
                 flags=Qt.WindowType.Dialog, **kwargs):
        super().__init__(parent, flags=flags, **kwargs)

        self._parent = parent
        self._getter = getter

        self._dataObject_copy = copy (self.dataObject)

        if width is not None: 
            self._width = width
        if height is not None: 
            self._height = height

        # set width and height 

        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        # move self relative to parent bottom left
        #       calc from the relative parentPos and dialogPos arguments the absolute positioning

        parent_topLeft     = self._parent.mapToGlobal (self._parent.rect().topLeft())
        parent_bottomRight = self._parent.mapToGlobal (self._parent.rect().bottomRight())
        parent_width  = parent_bottomRight.x() - parent_topLeft.x()
        parent_height = parent_bottomRight.y() - parent_topLeft.y()

        parentPos_rel = QPoint (int(parent_width * parentPos[0]), int(parent_height * parentPos[1]))
        parentPos_abs = parent_topLeft + parentPos_rel

        self_width  = self.rect().width()
        self_height = self.rect().height()
        selfPos_rel = QPoint (int(self_width * dialogPos[0]), int(self_height * dialogPos[1]))

        if dx is not None and dy is not None:
            dxy_rel = QPoint (int(dx), int(dy))
        else:
            dxy_rel = QPoint (0,0)

        self.move (parentPos_abs - selfPos_rel + dxy_rel)

        # title of dialog 

        if title is not None: 
            self.name = title 
        self.setWindowTitle (self.name)

        # enable custom window hint, disable (but not hide) close button
        # self.setWindowFlag(Qt.WindowType.WindowCloseButtonHint, False)

        # inital content panel content - layout in _init_layout()  

        self._panel = QWidget () 
        self.set_background_color (darker_factor=105)

        l_panel = self._init_layout()                               # subclass will create layout 
        l_panel.setContentsMargins (QMargins(15, 10, 15, 10))       # inset left 
        self._panel.setLayout (l_panel)

        # Qt buttonBox at footer

        button_box = self._button_box()
        if button_box:
            l_button = QHBoxLayout()
            l_button.addWidget(self._button_box())
            l_button.setContentsMargins (QMargins(5, 0, 25, 0))
 
        # main layout with title and panel 

        l_main   = QVBoxLayout()
        l_main.addWidget (self._panel, stretch=1)
        if button_box:
            l_main.addLayout (l_button)
        l_main.setContentsMargins (QMargins(5, 5, 5, 15))
        l_main.setSpacing(15)
        self.setLayout (l_main)

        # connect to change signal of widget 
        
        for w in self.widgets:
            w.sig_changed.connect (self._on_widget_changed)


    def set_background_color (self, darker_factor : int | None = None,
                                    color : QColor | int | None  = None,
                                    alpha : float | None = None):
        """ 
        Set background color of a QWidget either by
            - darker_factor > 100  
            - color: QColor or string for new color
            - alpha: transparency 0..1 
        """
        set_background (self._panel, darker_factor=darker_factor, color=color, alpha=alpha)


    def _init_layout(self) -> QLayout:
        """ init and return main layout"""

        # to be implemented by sub class
        return QVBoxLayout ()

    def _button_box (self) -> QDialogButtonBox:
        """ returns the QButtonBox with the buttons of self"""
        buttons = QDialogButtonBox.StandardButton.Ok | \
                  QDialogButtonBox.StandardButton.Cancel
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        return buttonBox 

    def _QButtons (self):
        """return QButtons enum for button box at footer """
        # to overload 
        return QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel


    @property
    def dataObject (self): 
        """ dataObject the dialog got from parent """
        # to be overloaded - or implemented with semantic name 
        if callable(self._getter):
            return self._getter()
        else: 
            return self._getter

    @property
    def dataObject_copy (self): 
        """ shallow copy of dataObject"""
        # to be overloaded - or implemented with semantic name 
        return self._dataObject_copy

    @property
    def widgets (self) -> list[Widget]:
        """ list of widgets defined in self """
        return Panel_Abstract.widgets_of_layout (self._panel.layout())


    def _on_widget_changed (self,*_):
        """ slot for change of widgets"""
        # to be overloaded 
        pass


    def button_clicked (self, aButton): 
        """ slot for button of buttonbox clicked. Can be overridden"""
        pass


    def refresh(self, disable=None):
        """ refreshes all Widgets on self """

        for w in self.widgets:
            w.refresh(disable=disable)
        logger.debug (f"{self} - refresh")


    @override
    def reject (self): 
        """ handle reject (close) actions"""
        # to override 
        super().reject()


#-------------------------------------------------------------------------------
# Tab panel    
#-------------------------------------------------------------------------------


class Tab_Panel (QTabWidget):
    """ 
    Tab Widget as parent for other items 
    """

    name = "Panel"             # will be title 

    _width  = None
    _height = None 


    def __init__(self,  
                 width=None, 
                 height=None, 
                 **kwargs):
        super().__init__(**kwargs)

        if width  is not None: self._width = width
        if height is not None: self._height = height

        # set width and height 
        Widget._set_width  (self, self._width)
        Widget._set_height (self, self._height)

        font = self.font() 
        _font = size.HEADER.value
        font.setPointSize(_font[0])
        font.setWeight   (_font[1])  
        self.setFont(font)

        # see https://doc.qt.io/qt-6/stylesheet-examples.html

        if Widget.light_mode:
            tab_style = """
            QTabWidget::pane { /* The tab widget frame */
                border-top:1px solid #ababab;
            }

            QTabWidget::tab-bar {
                left: 400px; /* move to the right by 5px */
            }

            /* Style the tab using the tab sub-control. Note that
                it reads QTabBar _not_ QTabWidget */
            QTabBar::tab {
                /*background: green; */
                border: 1px solid #C4C4C3;
                border-bottom: 0px;                                     /*remove */
                border-top-left-radius: 3px;
                border-top-right-radius: 3px;
                min-width: 25ex;
                padding: 6px;
            }

            QTabBar::tab:!selected {
                margin-top: 2px; /* make non-selected tabs look smaller */
                background: #e5e5e5
            }
                            
            QTabBar::tab:hover {
                background: rgba(255, 255, 255, 0.2) /* rgba(255, 20, 147, 0.1); */              
            }

            QTabBar::tab:selected {
                background: #000000 /*rgba(255, 255, 255, 0.9) */;               
            }

            QTabBar::tab:selected {
                color: #E0E0E0 /* #303030*/;
                font-weight: 600;
                border-color: #9B9B9B;
                /*border-bottom-color: #C2C7CB;  */
                border-bottom-color: red; /* same as pane color */
            }
            """
 
        else: 

            tab_style = """
            QTabWidget::pane { /* The tab widget frame */
                border-top:1px solid #505050;
            }

            QTabWidget::tab-bar {
                left: 400px; /* move to the right by 5px */
            }

            /* Style the tab using the tab sub-control. Note that
                it reads QTabBar _not_ QTabWidget */
            QTabBar::tab {
                /*background: green; */
                border: 1px solid #505050;  
                border-bottom: 0px;                                     /*remove */
                border-top-left-radius: 3px;
                border-top-right-radius: 3px;
                min-width: 25ex;
                padding: 6px;
            }

            QTabBar::tab:!selected {
                margin-top: 2px; /* make non-selected tabs look smaller */
                color: #D0D0D0;
                background: #353535
            }
                            
            QTabBar::tab:hover {
                background: rgba(255, 255, 255, 0.2) /* rgba(255, 20, 147, 0.1); */             
            }

            QTabBar::tab:selected {
                background: rgba(77, 77, 77, 0.9) /* background: rgba(255, 20, 147, 0.2); */                   
            }

            QTabBar::tab:selected {
                /*color: white; */
                color: #E0E0E0;
                font-weight: 600;
                border-color: #909090;
                border-bottom-color: #C2C7CB;   /* same as pane color */
            }
            """


        self.setStyleSheet (tab_style) 


    def __repr__(self) -> str:
        # overwritten to get a nice print string 
        return f"<Tab_Panel '{self.name}'>"


    def add_tab (self, aWidget : QWidget, name : str = None):
        """ at an item having 'name' to self"""

        if name is None:
            try:
                name = aWidget.name
            except AttributeError:
                name = aWidget.__class__.__name__

        self.addTab (aWidget, name)


    @property
    def current_tab_name (self) -> str:
        """ return the label of the current tab """
        return self.tabText (self.currentIndex())

    def set_current_tab (self, name : str):
        """ set the current tab to tab with label 'name' """
        if name:
            for itab in range (self.count()):
                if self.tabText(itab) == name:
                    self.setCurrentIndex (itab)
                    return
        else:
            self.setCurrentIndex (0)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Resource management helper

Supports both pip-installed packages and PyInstaller frozen executables.
"""

import sys
from pathlib import Path

# subdirectory for resources in PyInstaller frozen executable onedir mode
FROZEN_RESOURCES_SUBDIR   = '_internal/airfoileditor'

# directory for xo2 examples relative to resources directory
XO2_EXAMPLE_DIR             = "examples_optimize"


def _is_frozen() -> bool:
    """Check if running as PyInstaller frozen executable."""
    return getattr(sys, 'frozen', False)


def _get_resources_dir_pyinstaller() -> Path:
    """Get resources directory for PyInstaller frozen executable."""
    # PyInstaller bundles to _internal directory
    exe_dir = Path(sys.executable).parent
    return exe_dir  / FROZEN_RESOURCES_SUBDIR


def _get_resources_dir_pip() -> Path:
    """ Get resources directory for pip-installed package."""
    return Path(__file__).parent


def _get_resources_dir_dev() -> Path:
    """ Get resources directory for local development"""
    return Path(__file__).parent.parent



def get_resource_path(resource_type: str, *path_parts: str) -> Path:
    """
    Get path to a resource (icon, template, example).
    
    Works for pip packages, dev mode and PyInstaller executables.
    
    Args:
        resource_type: 'icons', 'templates', or 'examples'
        *path_parts: Additional path components (e.g., 'PC2.ico')
    
    Returns:
        Path to the resource
        
    Examples:
        >>> get_resource_path('icons', 'PC2.ico')
        >>> get_resource_path('examples', 'VJX.glide', 'VJX.glide.pc2')
    """
    if _is_frozen():
        # PyInstaller: resources are in _internal directory
        base = _get_resources_dir_pyinstaller()
    else:
        # Pip package: resources are at project root level
        # (icons/, templates/, examples/ next to planformcreator2/)
        base = _get_resources_dir_pip()

        if not (base / resource_type).exists():
            # Development mode: resources are in parent directory
            base = _get_resources_dir_dev()
    
    resource_path = base / resource_type / Path(*path_parts) if path_parts else base / resource_type

    return resource_path


def get_icon_path(icon_file_name: str) -> Path:
    """
    Get path to an icon file.
    """
    if icon_file_name:
        icon_path = get_resource_path('icons', icon_file_name)
        if icon_path.is_file():
            return icon_path
    return None


def get_icons_path() -> Path:
    """Get the icons directory path."""
    return get_resource_path('icons')


def get_template_dir() -> Path:
    """Get the templates directory path."""
    return get_resource_path('templates')


def get_examples_dir() -> Path:
    """Get the examples directory path."""
    return get_resource_path('examples')


def get_assets_dir() -> Path:
    """Get the assets directory path."""
    return get_resource_path('assets')


def get_xo2_examples_dir() -> Path:
    """Get the xo2 examples directory path."""
    return get_resource_path('examples_optimize')


def get_icon_as_string(icon_name: str) -> str:
    """
    Get icon path as string for compatibility with existing code.
    
    Args:
        icon_name: Icon filename
        
    Returns:
        Path as string
    """
    return str(get_icon_path(icon_name))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Proxies to Xoptfoil2 and Worker 

If the programs aren't in path-environment or in 'assets directory'
the path to program location location must be set 

   Worker.exePath = "xy/z/"
"""


import os
from tempfile               import NamedTemporaryFile
from glob                   import glob
from pathlib                import Path
from packaging.version      import Version                                  # has to be installed

import time 
import shutil
import logging
import datetime 
import fnmatch

from subprocess             import Popen, run, PIPE
if os.name == 'nt':                                 # startupinfo only available in windows environment  
    from subprocess import STARTUPINFO, CREATE_NEW_CONSOLE, STARTF_USESHOWWINDOW, CREATE_NO_WINDOW


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


SW_NORMAL = 1 
SW_MINIMIZE = 6 

EXE_DIR_WIN    = 'windows'                          # directory of exe files 
EXE_DIR_UNIX   = 'linux'                    

TMP_INPUT_NAME = 'tmp~'                             # temporary input file (~1 will be appended)
TMP_INPUT_EXT  = '.inp'

#------- Helper function -----------------------------------------#

def is_younger (filePath, age_in_seconds): 
    """ returns true if filePathName is younger than age_in_seconds """

    younger = False 

    if os.path.isfile (filePath):
        now         = datetime.datetime.now()
        last_update = datetime.datetime.fromtimestamp(os.path.getmtime(filePath))
        tdelta = now - last_update
        seconds = tdelta.total_seconds()
        younger =  seconds < age_in_seconds

    return younger


def file_in_use (filePath):
    """ returns True if file is in use by another process"""
    
    in_use = False

    if os.path.exists(filePath):
        try:
            os.rename(filePath, filePath)
        except OSError as e:
            logger.warning (f"File {filePath} in use by another process")
            in_use = True 

    return in_use 



#------------------------------------------------------------------------------------

class X_Program:
    """ 
    Abstract superclass - Proxy to execute eg Xoptfoil2 and Worker 
    
        self will be executed in 'workingDir' which must be set 
            if it can't be extracted from airfoil path
    """

    NAME        = 'My_Program'
    NAME_EXE    = 'my_program'                             # stem of of exe file 

    version     = ''                                       # version of self - will be set in isReady
    exe_dir     = None                                     # where to find .exe 
    ready       = False                                    # is Worker ready to work 
    ready_msg   = ''                                       # ready or error message 


    def __init__ (self, workingDir : str = None):

        if workingDir and not os.path.isdir (workingDir):
            raise ValueError (f"Working directory '{workingDir}' does not exist" )

        self._workingDir        = workingDir                # directory in which self will be executed
        self._popen : Popen     = None                      # instance of subprocess when async
        self._returncode        = 0                         # returncode when async finished 
        self._pipe_error_lines  = None                      # errortext lines from stderr when finished
        self._pipe_out_lines    = []                        # output lines from stdout when finished

        self._tmp_inpFile       = None                      # tmpfile of this instance, to be deleted 


    def __repr__(self) -> str:
        """ nice representation of self """
        return f"<{type(self).__name__}>"


    @property
    def workingDir (self) -> str: 
        """ directory in which self will be executed"""
        return self._workingDir
    
   
    def isReady (self, assets_dir : str, min_version : str = '') -> bool:
        """ 
        checks if self is available with min_version.

        Args: 
            assets_dir: directory where there should be ./windows or ./linux subdir with exe 
            min_version: check fpr min version number 
        """

        # ready already checked? 

        if self.ready: return True

        # find .exe

        version_ok = False
        ready_msg  = None 
        cls = self.__class__

        if self.exe_dir is None: 

            exe_dir, ready_msg = self._get_exe_dir (assets_dir)

            if exe_dir is None:                                        # self not found anywhere
                cls.ready_msg = ready_msg
                logger.error (ready_msg)
                return 
            else:     
                cls.exe_dir = exe_dir
                logger.debug (ready_msg)

        # try to execute with -h help argument to get version 

        returncode = self._execute ('-h', workingDir=self._get_workingDir(), capture_output=True)


        # extract version and check version
        if returncode == 0 :
            for line in self._pipe_out_lines: 
                words = line.split()

                # first word is program name 
                if len (words) > 1 and words[0] == self.NAME:

                    # last word is version - compare single version numbers
                    cls.version = words[-1]
                    version_ok = Version (cls.version) >= Version(min_version) if min_version else True
            
            if not version_ok:
                cls.ready_msg = f"wrong version {self.version} - need {min_version}"
            else: 
                cls.ready = True
                cls.ready_msg = f"Ready"

        else: 
            cls.ready_msg = f"{self.NAME_EXE} couldn't be executed"      

        if self.ready: 
            logger.info (f"{self.NAME} {self.version} {self.ready_msg}  (loading from: {self.exe_dir})" )
        else: 
            logger.error (f"{self.ready_msg}  (loading from: {self.exe_dir})" )

        return self.ready
    

    def isRunning (self) -> bool:
        """ still running when async - otherwise False"""

        isRunning = False
        if self._popen is not None: 

            self._popen.poll ()                 # check return code

            if self._popen.returncode is None: 

                logger.debug (f"... {self.NAME_EXE} running")
                isRunning = True

            else: 

                self._returncode       = self._popen.returncode

                new_lines = self._popen.stderr.readlines() if self._popen.stderr else []
                new_lines = [x.rstrip() for x in new_lines]                 # remove \r
                self._pipe_error_lines.extend(new_lines)
                
                new_lines = self._popen.stdout.readlines() if self._popen.stdout else []
                new_lines = [x.rstrip() for x in new_lines]
                self._pipe_out_lines.extend(new_lines)

                if self._returncode:
                    logger.error (f"... {self.NAME_EXE} returncode: {self._returncode} - {'\n'.join (self._pipe_error_lines)}")

                    # put minimum info in error_lines if it isn't standard error return code of Xoptfoil
                    if self._returncode > 1: 
                        self._pipe_error_lines = [f"Process error: {self._returncode}"]

                else: 
                    logger.debug (f"... {self.NAME_EXE} finished {'\n'.join (self._pipe_out_lines)}")
                 
                self._popen = None              # close down process instance 

        return isRunning


    def remove_tmp_file (self, iretry = 0):
        """ os remove of temporary input file"""

        if self._tmp_inpFile and os.path.isfile (self._tmp_inpFile):       # remove tmpfile of worker 
            try: 

                os.remove(self._tmp_inpFile) 
                self._tmp_inpFile = None 

            except OSError as exc: 
                if iretry < 5: 
                    iretry +=1
                    logger.warning (f"{self} Could not delete tmp input file '{self._tmp_inpFile}' - Retry {iretry}")
                    time.sleep (0.1)
                    self.remove_tmp_file (iretry=iretry)
                else: 
                    logger.error (f"Could not delete tmp input file '{self._tmp_inpFile}' - {exc}")
        else:
            self._tmp_inpFile = None


    @property
    def finished_returncode (self): 
        """ returncode of subprocess when finished"""
        return self._returncode


    @property
    def finished_errortext (self): 
        """ errortext from subprocess in case returncode !=0 """

        # scan stderr and stdout for a line with 'Error: ' 

        text = None
        line : str

        if self._pipe_error_lines:
            for line in self._pipe_error_lines:
                _,_,text = line.partition ("Error: ")
                if text: return text 

            # if process was aborted no piped error text from xoptfoil2 - but a minimum message 
            return self._pipe_error_lines[0]

        if self._pipe_out_lines:
            for line in self._pipe_out_lines:
                _,_,text = line.partition ("Error: ")
                if text: return text 

        return text


    def finalize (self):
        """ do cleanup actions """

        self.remove_tmp_file ()


    def terminate (self):
        """ terminate os process of self"""

        if self.isRunning():
            self._popen.terminate()
            self._popen = None 


    # ---------- Private --------------------------------------

    def _execute (self, args = [], workingDir = None, capture_output : bool =False):
        """sync execute self in workingDir 

        Args:
            args: arguments of subprocess as list of strings
            capture_output: capture output in pipe. Defaults to False.
        Returns:
            returncode: = 0 if no error
        """

        if workingDir and not os.path.isdir (workingDir):
            returncode = 1
            logger.error (f"Working directory '{workingDir}' does not exist" )
            return returncode

        returncode  = 0 
        self._pipe_out_lines    = []                        # output lines from stderr when finished
        self._pipe_error_lines  = []                        # errortext lines from stdout when finished

        # build list of args needed by subprocess.run 

        exe = os.path.join (self.exe_dir, self.NAME_EXE) 

        if isinstance (args, list):
            arg_list = [exe] + args
        elif isinstance (args, str):
            arg_list = [exe] + [args]
        else: 
            arg_list = [exe]

        try:

            # uses subproocess run which returns a completed process instance 

            curDir = os.getcwd()
            if workingDir:
                os.chdir (workingDir)

            if capture_output:
                # needed when running as pyinstaller .exe 
                # https://stackoverflow.com/questions/7006238/how-do-i-hide-the-console-when-i-use-os-system-or-subprocess-call/7006424#7006424
                if os.name == 'nt':
                    flags = CREATE_NO_WINDOW    
                else: 
                    flags  = 0                      # posix must be 0       

            logger.info (f"... {self.NAME_EXE} run sync: '{" ".join(args)}' in: {workingDir}")

            process = run (arg_list, text=True, capture_output=capture_output, creationflags=flags)

            returncode  = process.returncode

            if returncode:
                logger.error (f"... {self.NAME_EXE} ended: '{process}'")

            # finished - nice output strings 

            if process.stderr:  
                self._pipe_error_lines = process.stderr.split ("\n")
                logger.error (f"... {self.NAME_EXE} stderr: {"\n".join (self._pipe_error_lines)}")

                if self._pipe_error_lines[0] == "STOP 1":
                    # the error message will be in stdout
                    self._pipe_error_lines = []

            if capture_output and process.stdout: 
                self._pipe_out_lines = process.stdout.split ("\n")
                # logger.debug (f"... {self.NAME_EXE} stdout: {"\n".join (self._pipe_out_lines)}")

        except FileNotFoundError as exc:

            returncode = 1
            self._pipe_error_lines = str(exc)

            logger.error (f"... exception {self.NAME_EXE}: {exc}")

        finally: 

            os.chdir (curDir)

        return  returncode


    def _execute_async (self, args = [], workingDir = None, capture_output : bool =False):
        """async execute self in workingDir 

        Args:
            args: arguments of subprocess as list of strings
            capture_output: capture output in pipe. Defaults to False.
        Returns:
            returncode: = 0 if no error
        """

        if workingDir and not os.path.isdir (workingDir):
            raise ValueError (f"Working directory '{workingDir}' does not exist" )

        returncode  = 0 
        self._pipe_out_lines    = []                        # errortext lines from stderr when finished
        self._pipe_error_lines  = []                        # output lines from stdout when finished

        # build list of args needed by subprocess.run 

        exe = os.path.join (self.exe_dir, self.NAME_EXE) 

        if isinstance (args, list):
            arg_list = [exe] + args
        elif isinstance (args, str):
            arg_list = [exe] + [args]
        else: 
            arg_list = [exe]
        # run either sync or async 

        try:

            # uses subproccess Popen instance to start a subprocess

            curDir = os.getcwd()
            if workingDir:
                os.chdir (workingDir)

            if capture_output:
                stdout = PIPE                               # output is piped to suppress window 
                stderr = PIPE                               # Xoptfoil will write error to stderr
                # needed when running as pyinstaller .exe 
                # https://stackoverflow.com/questions/7006238/how-do-i-hide-the-console-when-i-use-os-system-or-subprocess-call/7006424#7006424
                if os.name == 'nt':
                    flags = CREATE_NO_WINDOW    
                else: 
                    flags  = 0                              # posix must be 0 
                startupinfo = self._get_popen_startupinfo (SW_NORMAL)  
            else: 
                stdout = None 
                stderr = PIPE                               # Xoptfoil will write error to stderr
                if os.name == 'nt':
                    flags  = CREATE_NEW_CONSOLE             # a new console is created (Xoptfoil) 
                else: 
                    flags  = 0                              # posix must be 0 
                startupinfo = self._get_popen_startupinfo (SW_MINIMIZE)  

            logger.info (f"... {self.NAME_EXE} run async: '{" ".join(args)}' in: {workingDir}")

            popen = Popen (arg_list, creationflags=flags, text=True, **startupinfo, 
                                stdout=stdout, stderr=stderr)  

            popen.poll()                            # update returncode

            returncode  = popen.returncode if popen.returncode is not None else 0   # async returns None 

            if returncode:
                logger.error (f"... {self.NAME_EXE} ended: '{popen}'")

                self._pipe_error_lines = popen.stderr.readlines()
                logger.error (f"... {self.NAME_EXE} stderr: {"\n".join (self._pipe_error_lines)}")

            # keep for later poll 
            self._popen = popen 

        except FileNotFoundError as exc:

            returncode = 1
            self._pipe_error_lines = str(exc)

            logger.error (f"... exception {self.NAME_EXE}: {exc}")

        finally: 

            os.chdir (curDir)

        return  returncode



    def _get_popen_startupinfo (self, show : int):
        """ returns popen startinfo parm to eg. minimize shell window - only windows"""

        if os.name == 'nt':
            if show != SW_NORMAL:
                startupinfo = STARTUPINFO()  
                startupinfo.dwFlags |= STARTF_USESHOWWINDOW       
                startupinfo.wShowWindow = show
                return dict(startupinfo=startupinfo)
        return dict(startupinfo=None) 


    def _get_exe_dir (self, assets_dir : str): 
        """
        trys to find path to call programName
        
        If found, returns exePath and ready_msg
        If not, return None and ready_msg (error)"""

        exe_dir  = None
        ready_msg = None 

        if os.name == 'nt':
            exe_dir = EXE_DIR_WIN
        else: 
            exe_dir = EXE_DIR_UNIX  

        check_dir = os.path.join (assets_dir , exe_dir)                            #

        if shutil.which (self.NAME_EXE, path=check_dir) : 
            exe_dir  = os.path.abspath(check_dir) 
            ready_msg = f"{self.NAME_EXE} found in: {exe_dir}"
        else: 
            exe_path = shutil.which (self.NAME_EXE)  
            if exe_path: 
                exe_dir = os.path.dirname (exe_path)
                ready_msg = f"{self.NAME_EXE} using OS search path to execute: {exe_dir}"
            else: 
                exe_dir = None
                ready_msg = f"{self.NAME_EXE} not found either via OS search path nor in '{check_dir}'" 
        return exe_dir, ready_msg


    def _get_workingDir (self, airfoil_path : str = None) -> str:
        """ 
        returns the actual working dir in which X_Program will be executed
            A workingDir set for self will have presetence, otherwise it is
            retrieved from airfoil_path or os current working dir is taken  
        """

        if self.workingDir:
            return self.workingDir
        elif airfoil_path and os.path.isfile (airfoil_path) : # and os.path.isabs(airfoil_path)
            return os.path.dirname (airfoil_path)
        else:
            return os.getcwd()



# ------------------------------------------------------------



class Xoptfoil2 (X_Program):
    """ 
    Proxy to execute Xoptfoil2
    
        self will be executed in 'workingDir' which must be set if is not current dir
        The 'inputfile' must be in 'workingDir' 
    """

    NAME        = 'Xoptfoil2'
    NAME_EXE    = 'xoptfoil2'                               # stem of of exe file 

    RUN_CONTROL = 'run_control'                             # file name of control file 
    STILL_ALIVE = 10                                        # max. age in seconds of run_control

    RESULT_DIR_POSTFIX = '_temp'                            # result directory of Xoptfoil2 postfix of 'outname'

    @staticmethod
    def remove_resultDir (airfoil_pathFileName : str, only_if_older = False):
        """ 
        deletes the Xoptfoil2 result directory 
        """ 

        resultDir = str(Path(airfoil_pathFileName).with_suffix('')) + Xoptfoil2.RESULT_DIR_POSTFIX
        shutil.rmtree(resultDir, ignore_errors=True)


    @property
    def run_control_filePath (self):
        """ returns filePath of run_control"""
        if self.workingDir: 
            return os.path.join(self.workingDir, Xoptfoil2.RUN_CONTROL)
        else:
            return Xoptfoil2.RUN_CONTROL


    def run (self, outname:str, input_file:str=None, seed_airfoil:str =None):
        """ run self async in self workingDir

        Args:
            outname: output name for generated airfoil
            inputfile: name of input file. Defaults to 'outname'.inp.
            seed_airfoil: optional seed airfoil filename.
        Returns: 
            returncode: = 0 - no errors (which could be retrieved via 'finished_errortext' )
        """

        args = []
        if seed_airfoil  : args.extend(['-a', seed_airfoil])
        if outname       : args.extend(['-o', outname])  
        
        if input_file is None: input_file = outname + '.inp'
        args.extend(['-i', input_file]) 

        # add 'mode' option - will write error to stderr
        args.extend(['-m', 'child']) 

        returncode = self._execute_async (args=args, workingDir=self.workingDir)

        return returncode


    def isRunning (self) -> bool:
        """ 
        - still running?    ... process when async
                            ... otherwise check run_control file if self was started from outside 
        """

        if self._popen:                         # started program myself as process
            running = super().isRunning ()
        else:                                   # Xoptfoil was started from outside 
            running = is_younger (self.run_control_filePath,  Xoptfoil2.STILL_ALIVE)

        if not running and os.path.isfile (self.run_control_filePath):
            # remove old run_control 
            os.remove (self.run_control_filePath)
 
        return running 



    def get_progress (self):
        """ returns no of steps, no of designs and objective function when running otherwise 0,0, 1.0 """

        # format of run_control 
            # !stop
            # !run-info; step: 3; design: 3; fmin:  0.9919478

        steps   = 0 
        designs = 0 
        objFun  = 1.0 
        lines = []

        if os.path.isfile (self.run_control_filePath):
            with open(self.run_control_filePath, 'r') as file:
                lines = file.readlines()
                file.close()

        for line in lines: 
            infos = line.split(";")
            if len(infos) == 4: 
                try: 
                    steps   = int(infos[1].split(":")[1])
                    designs = int(infos[2].split(":")[1])
                    objFun  = float(infos[3].split(":")[1])
                except: 
                    pass

        return steps, designs, objFun


    def stop (self):
        """
        tries to stop self with a 'stop' command in 'run_control' file 
            self must run in 'workingDir'
        """

        with open(self.run_control_filePath, 'w+') as file:
            file.write("stop")
            file.close()


# ------------------------------------------------------------



class Worker (X_Program):
    """ proxy to execute Worker commands"""

    NAME        = 'Worker'
    NAME_EXE    = 'worker'                             # stem of of exe file 


    # -- static methods --------------------------------------------

    @classmethod
    def can_detect_bubbles (cls) -> bool:
        """ 
        returns True if Worker version can detect bubbles in polars
        """
         
        return Version(cls.version) >= Version('1.0.9') if cls.version else False


    @staticmethod
    def polarDir (airfoil_pathFileName : str) -> str:
        """ returns polar directory of airfoil having airfoil_pathFileName"""

        return str(Path(airfoil_pathFileName).with_suffix('')) + '_polars'

    @staticmethod
    def flapped_suffix (flap_angle:float, x_flap:float, y_flap:float, y_flap_spec:str) -> str:
        """ 
        name extension for flapped airfoil or polar file 
            '_f5.1' for defaults or 
            '_f-1.4_xf0.72_yf0.5_yspecYC' for non default values
        """
        if flap_angle == 0.0: return ''

        s_flap_angle = f"_f{flap_angle:.2f}".rstrip('0').rstrip('.') 
        s_x_flap     = f"_xf{x_flap:.2f}".rstrip('0').rstrip('.') if x_flap != 0.75 else ''
        s_y_flap     = f"_yf{y_flap:.2f}".rstrip('0').rstrip('.') if y_flap != 0.0  else ''
        s_y_spec     = f"_yspecYC" if y_flap_spec == 'y/c'  else ''

        return  f"{s_flap_angle}{s_x_flap}{s_y_flap}{s_y_spec}"


    @staticmethod
    def remove_polarDir (airfoil_pathFileName : str, only_if_older = False):
        """ 
        deletes polar directory 
        If only_if_older the directory is not removed if it is younger than of airfoilPathFileName
        """ 

        polarDir = Worker.polarDir (airfoil_pathFileName) if airfoil_pathFileName else None

        # sanity check 
        if not os.path.isdir(polarDir): return 

        remove = True 

        if only_if_older:
            if os.path.isfile(airfoil_pathFileName):

                # compare datetime of airfoil file and polar dir 
                ts = os.path.getmtime(polarDir)                 # file modification timestamp of a file
                polarDir_dt = datetime.datetime.fromtimestamp(ts)        # convert timestamp into DateTime object

                ts = os.path.getmtime(airfoil_pathFileName)              # file modification timestamp of a file
                airfoil_dt = datetime.datetime.fromtimestamp(ts)         # convert timestamp into DateTime object

                # add safety seconds (async stuff?) 
                if (airfoil_dt < (polarDir_dt + datetime.timedelta(seconds=2))):
                    remove = False 
            else: 
                remove = False 

        if remove: 
            shutil.rmtree(polarDir, ignore_errors=True)


    @staticmethod
    def get_existingPolarFile (airfoil_pathFileName, 
                               polarType : str, re : float, ma : float, ncrit : float,
                               flap_angle : float, x_flap : float, y_flap : float, y_flap_spec : str) -> str:
        """ 
        Get pathFileName of polar file if it exists 
        """      

        def parm_is_ok (id:str, val : float|None, decimals, args :list[str]) -> bool:
            """ inner func: check if paramter is in argunents of fileName"""
            if val is not None and decimals is not None: 
                val = round (val, decimals)

            for arg in args:
                if id == arg [:len(id)]:
                    if val is not None:
                        arg_val = arg[len(id):] if isinstance (val, str) else float(arg[len(id):])
                        if val == arg_val:                  # right arg 
                            return True                     #   right value
                        else:
                            return False                    #   wrong value
                    else: 
                        return False                        # this arg is too much
            
            if val is None: 
                return True                                 # this arg is correctly not there 
            else:
                return False                                # this arg is missing


        # remove a maybe older polarDir
        Worker.remove_polarDir (airfoil_pathFileName, only_if_older=True)    

        # build name of polar dir from airfoil file 
        polarDir = Worker.polarDir (airfoil_pathFileName)
        if os.path.isdir (polarDir):         

            fileNames = fnmatch.filter(os.listdir(polarDir), '*.txt')
            for fileName in fileNames:

                args = Path(fileName).stem.split('_')

                ok = True 

                # classic part 'T1_Re0.500_M0.00_N7.0'

                ok = ok and parm_is_ok ("Re", re/1000000, 3, args)
                ok = ok and parm_is_ok ("M",  ma, 2, args)
                ok = ok and parm_is_ok ("N",  ncrit, 1, args)
                ok = ok and parm_is_ok ("T",  int(polarType[1:]), 0, args)

                # flapped part '_f-1.4_xf0.72_yf0.5_yspecYC' for non default values

                ok = ok and parm_is_ok ("f", flap_angle, 1, args)

                x_flap_arg = None if x_flap ==0.75 else x_flap
                ok = ok and parm_is_ok ("xf", x_flap_arg, 2, args)

                y_flap_arg = None if y_flap ==0.0 else y_flap
                ok = ok and parm_is_ok ("yf", y_flap_arg, 2, args)

                y_flap_spec_arg = 'YC' if y_flap_spec =='y/c' else None
                ok = ok and parm_is_ok ("yspec", y_flap_spec_arg, None, args)

                if ok:
                    # logger.debug (f"<class Worker> found polar file {fileName} in {polarDir}")
                    return os.path.join (polarDir, fileName)        # return pathFileName

        logger.debug (f"<class Worker> No polar file in {polarDir}")
        return None


    #---------------------------------------------------------------


    def check_inputFile (self, inputFile=None):
        """ uses Worker to check an Xoptfoil2"""

        if not self.ready: return 1, self.NAME + " not ready"

        error_text = ""
        args = ['-w', 'check-input', '-i', inputFile]

        returncode = self._execute (args, capture_output=True)

        if returncode != 0:

            # worker output should something like ...
            #  Worker   -check-input jx-gt-10v3.inp
            #  - Processing input
            #    - Reading input jx-gt-10v3.inp
            #    - Output prefix jx-gt-10v3
            #  Error: max_speed should be between 0.001 and 0.5

            for line in self._pipe_out_lines:
                error_text = line.partition("Error:")[2].strip()
                if error_text != '': break 
            if error_text == '':
                raise ValueError ("Errortext not found in Workers")

        return returncode, error_text 


    def generate_polar (self, airfoil_pathFileName, 
                        polarType : str, 
                        re : float | list, 
                        ma : float | list, 
                        ncrit : float,
                        autoRange = True, spec = 'alpha', valRange= [-3, 12, 0.25], 
                        flap_angle : float | list = 0.0, x_flap=0.75, y_flap=0, y_flap_spec='y/t',
                        nPoints : int = None, 
                        detect_bubble = True,
                        run_async = True) -> int:
        """ 
        Generate polar for airfoilPathFileName in directory of airfoil.
        Returncode = 0 if successfully started (async) or finish (sync)
        """ 

        if not os.path.isfile(airfoil_pathFileName): 
            name = airfoil_pathFileName if len(airfoil_pathFileName) <= 40 else "..." + airfoil_pathFileName[-35:]
            raise ValueError (f"Airfoil '{name}' does not exist")

        if (polarType == 'T2'):
            polarTypeNo = 2
        else: 
            polarTypeNo = 1

        if spec == 'alpha':
            spec_al = True
        else: 
            spec_al = False

        if not isinstance (re, list): re = [re]
        if not isinstance (ma, list): ma = [ma]
        if flap_angle:
            if not isinstance (flap_angle, list): flap_angle = [flap_angle]
        else: 
            flap_angle=None

        # working directory in which self will execute 

        workingDir = self._get_workingDir (airfoil_pathFileName)

        # a temporary input file for polar generation is created

        self._tmp_inpFile = self._generate_polar_inputFile (workingDir, 
                                    re, ma, polarTypeNo, ncrit, autoRange, spec_al, valRange,
                                    flap_angles=flap_angle, x_flap=x_flap, y_flap=y_flap, y_flap_spec=y_flap_spec,
                                    nPoints=nPoints, detect_bubble=detect_bubble) 
        if not self._tmp_inpFile:
            raise RuntimeError (f"{self.NAME} polar generation failed: Couldn't create input file")

        # build args for worker - force outname as Worker wrongly uses airfoil name!

        # outname = Path(os.path.basename(airfoil_pathFileName)).stem
        outname = '' 

        args = self._build_worker_args ('polar-flapped',airfoil1=airfoil_pathFileName, inputfile=self._tmp_inpFile, outname=outname)

        # .execute either sync or async

        if run_async: 

            returncode = self._execute_async (args, capture_output=True, workingDir=workingDir)

        else:

            returncode = self._execute       (args, capture_output=True, workingDir=workingDir)

            self.remove_tmp_file ()         
        
        if returncode: 
            raise RuntimeError (f"Worker polar generation failed for {airfoil_pathFileName}")
            



    def set_flap (self, airfoil_fileName, 
                        x_flap : float = 0.75,
                        y_flap : float = 0.0,
                        y_flap_spec : str = 'y/t',
                        flap_angle : float | list = 0.0,
                        outname : str =None) -> int:
        """ 
        Set flap for airfoilPathFileName in directory of airfoil.

        Returns fileName of flapped airfoil in working Dir  
        """ 

        # sanity check airfoil_file 

        if self.workingDir:
            airfoil_pathFileName_abs = os.path.join(self.workingDir, airfoil_fileName)
        else: 
            airfoil_pathFileName_abs = airfoil_fileName

        if not os.path.isfile(airfoil_pathFileName_abs): 
            name = airfoil_pathFileName_abs if len(airfoil_pathFileName_abs) <= 40 else "..." + airfoil_pathFileName_abs[-35:]
            raise ValueError (f"Airfoil '{name}' does not exist")

        # name of the generated airfoil needed

        if not outname: 
            outname=f"{Path(airfoil_fileName).stem}{Worker.flapped_suffix (flap_angle, x_flap, y_flap, y_flap_spec)}"

        # working directory in which self will execute 

        workingDir = self._get_workingDir (airfoil_fileName)

        # a temporary input file for polar generation is created

        self._tmp_inpFile = self._generate_flap_inputFile (workingDir, x_flap, y_flap, y_flap_spec, flap_angle)         
        if not self._tmp_inpFile:
            raise RuntimeError (f"{self.NAME} setting flap failed: Couldn't create input file")

        # build args for worker 

        args = self._build_worker_args ('flap', airfoil1=airfoil_fileName, 
                                                inputfile=self._tmp_inpFile,
                                                outname=outname)
        # .execute  sync

        rc = self._execute (args, capture_output=True, workingDir=workingDir)

        if rc: 
            raise RuntimeError (f"Worker set flap failed: {self.finished_errortext}")
        else: 
            self.remove_tmp_file ()         

            flapped_fileName = outname + ".dat"
            # check if flapped airfoil was created  
            if os.path.isfile (os.path.join (workingDir, flapped_fileName)):
                return flapped_fileName
            else: 
                return None 


    def clean_workingDir (self, workingDir):
        """ 
        deletes temporary (older) files Worker creates in workingDir
        """ 
        if os.path.isdir(workingDir):

            # remove tmp input files of polar generation 
            match_path = os.path.join (workingDir, f"{TMP_INPUT_NAME}*{TMP_INPUT_EXT}")
            
            for f in glob(match_path):
                os.remove(f)


# ---------- Private --------------------------------------



    def _build_worker_args (self, action, actionArg='', airfoil1='', airfoil2='', outname='', inputfile=''):
        """ 
        return worker args as list of strings  
        """

        airfoil1_fileName = os.path.basename(airfoil1)
        airfoil2_fileName = os.path.basename(airfoil2)

        # info inputfile is in a dir - strip dir from path - local execution 

        local_inputfile = os.path.basename(inputfile) if inputfile else ''

        args = []

        if (action != ''): 
            if (action == 'help'): 
                args.extend(['-h'])
            else:                  
                args.extend(['-w', action])
                if (actionArg): args.extend([actionArg]) 
        else: 
            raise ValueError ('action for worker is mandatory')

        if (airfoil1 ): args.extend(['-a',  airfoil1_fileName])
        if (airfoil2 ): args.extend(['-a2', airfoil2_fileName])
        if (outname  ): args.extend(['-o',  outname]) 
        if (inputfile): args.extend(['-i',  local_inputfile])

        # add 'mode' option - will write error to stderr
        args.extend(['-m', 'child']) 

        return  args



    def _generate_polar_inputFile (self, 
                                  workingDir : str,         # if none self._workingDir is taken 
                                  reNumbers : list[float], maNumbers : list[float],
                                  polarType : int, ncrit : float,  
                                  autoRange : bool, spec_al: bool, valRange: list[float], 
                                  nPoints : int = None, detect_bubble = False, 
                                  flap_angles : list[float] = None, x_flap=None, y_flap=None, y_flap_spec=None,
                                   ) -> str:
        """ Generate a temporary polar input file for worker like this 

        &polar_generation
            generate_polars = .true.
            polar_reynolds  = 230000, 300000, 400000
            polar_mach      = 0.0, 0.2, 0.5
            type_of_polar = 1
            auto_Range = .true.
            op_mode = 'spec-al'
            op_point_range = -2.6, 11.0, 0.5
        /
        &xfoil_run_options
            ncrit = 7.0
        /
        &operating_conditions                          ! options to describe the optimization task
        x_flap                 = 0.75                  ! chord position of flap 
        y_flap                 = 0.0                   ! vertical hinge position 
        y_flap_spec            = 'y/c'                 ! ... in chord unit or 'y/t' relative to height
        flap_angle             = 0.0                   ! list of flap angles to be set


        Returns: 
            pathFilename of input file  """

        tmpFilePath = None

        # create tmp input file 

        with NamedTemporaryFile(mode="w", delete=False, dir=workingDir, prefix=TMP_INPUT_NAME, suffix=TMP_INPUT_EXT) as tmp:

            tmp.write ("&polar_generation\n")
            tmp.write ("  type_of_polar = %d\n" % polarType)  
            tmp.write ("  polar_reynolds  = %s\n" % (', '.join(str(e) for e in reNumbers))) 
            if max(maNumbers) > 0.0: 
                tmp.write ("  polar_mach  = %s\n" % (', '.join(str(e) for e in maNumbers))) 
            if autoRange:
                tmp.write ("  auto_range = .true.\n") 
                if valRange[2]:                                 # write only increment
                    tmp.write ("  op_point_range = , , %.2f \n" % (valRange[2])) 
            else:
                if spec_al:  tmp.write ("  op_mode = 'spec-al'\n") 
                else:        tmp.write ("  op_mode = 'spec-cl'\n") 
                tmp.write ("  op_point_range = %.2f , %.2f , %.2f \n" % (valRange[0], valRange[1], valRange[2])) 
            tmp.write ("/\n")

            tmp.write ("&xfoil_run_options\n")
            tmp.write ("  ncrit = %.1f\n" % ncrit) 
            if detect_bubble and Worker.can_detect_bubbles():
                tmp.write ("  detect_bubble = .true.\n") 
            tmp.write ("/\n")

            if flap_angles:
                tmp.write ("&operating_conditions\n")
                tmp.write (f"  x_flap = {x_flap}\n")  
                tmp.write (f"  y_flap = {y_flap}\n")  
                tmp.write (f"  y_flap_spec = '{y_flap_spec}'\n")  
                tmp.write (f"  flap_angle = {', '.join(str(f) for f in flap_angles)}\n") 
                tmp.write ("/\n")

            if nPoints is not None: 
                tmp.write ("&paneling_options\n")
                tmp.write (f"  npoint = {int(nPoints)}\n")  
                tmp.write ("/\n")

            tmpFilePath = tmp.name

        return tmpFilePath              



    def _generate_flap_inputFile (self, 
                                 workingDir : str = None,         # if none self._workingDir is taken 
                                  x_flap : float = 0.75,
                                  y_flap : float = 0.0,
                                  y_flap_spec : str = 'y/c',
                                  flap_angle : float | list = 0.0, 
                                  ) -> str:
        """ Generate a temporary polar input file for worker inworkingDir

        &operating_conditions                              ! options to describe the optimization task
            x_flap                 = 0.75                  ! chord position of flap 
            y_flap                 = 0.0                   ! vertical hinge position 
            y_flap_spec            = 'y/c'                 ! ... in chord unit or 'y/t' relative to height
            flap_angle             = 0.0                   ! list of flap angles to be applied        
        /

        Returns: 
            pathFilename of input file  """

        tmpFilePath = None

        flap_angles = flap_angle if isinstance (flap_angle, list) else [flap_angle]
        y_flap_spec = 'y/t' if y_flap_spec == 'y/t' else 'y/c'

        # create tmp input file 

        with NamedTemporaryFile(mode="w", delete=False, dir=workingDir, prefix=TMP_INPUT_NAME, suffix=TMP_INPUT_EXT) as tmp:

            tmp.write ("&operating_conditions\n")
            tmp.write (f"  x_flap = {x_flap:.2f}\n")  
            tmp.write (f"  y_flap = {y_flap:.2f}\n")  
            tmp.write (f"  y_flap_spec = '{y_flap_spec}'\n")  
            tmp.write (f"  flap_angle = {','.join(map(str, flap_angles))}\n")  
            tmp.write ("/\n")

            tmpFilePath = tmp.name

        return tmpFilePath              




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Airfoil Editor 

    Object model overview (a little simplified) 

    App                                         - Main 
        |-- App_Mode                            - different UI modes like View, Modify, Optimize 
            |-- Data_Panel                      - UI lower data panel for mode 
                |-- Panel_Geometry              - UI single panel with fields 

            |-- Airfoil_Diagram                 - UI upper diagram area 
                |-- Airfoil_Diagram_Item        - UI a single plot item within diagram
                    |-- Curvature_Artist        - UI curvature plot artist
            
        |-- App_Model                           - App shell around data model allowing signals
            |-- Airfoil                         - airfoil model object 
            |-- Case                            - handle different like optimize, modify 
                |-- Xo2_Input                   - main object optimization - represents an X02 input file 
        
"""

from ast import Add
import os
import sys
import argparse

from PyQt6.QtCore           import pyqtSignal, QMargins, Qt
from PyQt6.QtWidgets        import QApplication, QMainWindow, QWidget 
from PyQt6.QtWidgets        import QGridLayout
from PyQt6.QtGui            import QCloseEvent, QGuiApplication, QIcon

# DEV: when running app.py as main, set package property to allow relative imports
if __name__ == "__main__":  
    sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    __package__ = "airfoileditor"

from .resources              import get_icons_path, get_icon_path
from .base.common_utils      import * 

from .model.xo2_input        import Input_File
from .model.case             import Case_Optimize, Case_Direct_Design

from .base.panels            import Win_Util
from .base.widgets           import Icon, Widget
from .base.app_utils         import Settings, Update_Checker, Run_Checker, check_or_get_initial_file

from .ui.ae_widgets          import create_airfoil_from_path
from .ui.ae_diagrams         import Diagram_Airfoil_Polar

from .app_model              import App_Model, Mode_Id
from .app_modes              import Modes_Manager, Mode_View, Mode_Modify, Mode_Optimize, Mode_As_Bezier

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------------------------
# The App   
#-------------------------------------------------------------------------------


APP_NAME         = "AirfoilEditor"
PACKAGE_NAME     = "airfoileditor"
__version__      = "4.2.1"                  # hatch "version dynamic" - PEP440 compliant version string
                                            # for Github use SemVer "4.2.0-beta.3"

CHANGE_TEXT      = "- Save / Load individual airfoil settings<br>" + \
                   "- Change polar diagram variables directly in diagram<br>" + \
                   "- Maximize / minimize lower data panel<br>" + \
                   "- Revised Match Bezier UI<br>"


class Main (QMainWindow):
    '''
        App - Main Window 
    '''

    # Qt Signals 

    sig_xo2_about_to_run        = pyqtSignal()          # short before optimization starts
    sig_xo2_new_state           = pyqtSignal()          # Xoptfoil2 new info/state


    def __init__(self, initial_file):
        super().__init__()

        logger.info (f"Init Main Window")

        # --- init Settings, check for newer app version ---------------

        Settings.set_file (APP_NAME, file_extension= '.settings')

        is_first_run = Run_Checker.is_first_run (__version__)                 # to show Welcome message
        
        Update_Checker (self, APP_NAME, PACKAGE_NAME,  __version__) 


        # --- init App Model ---------------

        app_model = App_Model (workingDir_default=Settings.user_data_dir (APP_NAME))

        app_model.set_app_info (__version__, CHANGE_TEXT, is_first_run)

        # either airfoil file or Xoptfoil2 input file

        initial_file = check_or_get_initial_file (initial_file)

        if Input_File.is_xo2_input (initial_file):
            initial = initial_file
            mode_to_start = Mode_Id.OPTIMIZE
        else:
            airfoil = create_airfoil_from_path (self, initial_file, example_if_none=True, message_delayed=True)
            initial = airfoil
            mode_to_start = Mode_Id.VIEW

        # load initial settings like polar definitions 

        app_model.load_settings ()                                          # global settings (no airfoil set)
        app_model.sig_new_airfoil.connect (self._set_win_title)             # update title on new airfoil
        app_model.sig_new_mode.connect    (self._set_win_title)             # update title on new case

        self._app_model     = app_model                                     # keep for close 


        # --- init UI ---------------

        # main window style - dark or light mode

        self._set_win_style (app_icon_name='AE.ico')
        self._set_win_title ()
        self._set_win_geometry ()

        # app Modes and manager  
        
        logger.info (f"Init Modes Manager")

        modes_manager = Modes_Manager (app_model)
        modes_manager.add_mode (Mode_View       (app_model))
        modes_manager.add_mode (Mode_Modify     (app_model))
        modes_manager.add_mode (Mode_Optimize   (app_model))
        modes_manager.add_mode (Mode_As_Bezier  (app_model))

        modes_manager.set_mode (mode_to_start, initial)                     # set initial object in app_model
        modes_manager.sig_close_requested.connect (self.close)              # app close requested from mode view

        self._modes_manager = modes_manager                                 # keep as it hosts slots

        # main widgets and layout of app

        logger.info (f"Init UI - mode: {modes_manager.current_mode}")

        diagram     = Diagram_Airfoil_Polar (self,app_model)                # big diagram widget
        modes_panel = modes_manager.stacked_modes_panel()                   # stacked widget with mode data panels

        l = QGridLayout () 
        l.addWidget (diagram,     0,0)
        l.addWidget (modes_panel, 1,0)

        l.setRowStretch (0,1)
        l.setRowMinimumHeight (0,400)
        modes_manager.set_height (180, minimized=65)
        
        l.setSpacing (5)
        l.setContentsMargins (QMargins(5, 5, 5, 5))

        main = QWidget()
        main.setLayout (l) 
        self.setCentralWidget (main)

        self._diagram = diagram                                             # keep for close down


        # --- Enter event loop ---------------

        logger.info (f"{modes_manager.current_mode} ready - entering event loop")


    # --- private ---------------------------------------------------------


    def _set_win_title (self):
        """ set window title with airfoil or case name """

        case    = self._app_model.case
        airfoil = self._app_model.airfoil
        if isinstance (case, Case_Optimize):
            ext = f"[Case {case.name if case else '?'}]"
        elif isinstance (case, Case_Direct_Design):
            seed   = case.airfoil_seed.fileName  
            design = airfoil.name_to_show
            ext = f"[{design} on {seed}]"
        else: 
            ext = f"[{airfoil.fileName if airfoil else '?'}]"

        self.setWindowTitle (APP_NAME + "  v" + str(__version__) + "  " + ext)


    def _set_win_style (self, icons_path : Path = None, app_icon_name : str  = None):
        """ 
        Set window style according to settings
        """

        # set resources dir for Icons
        if icons_path is None:
            icons_path = get_icons_path()
        Icon.ICONS_PATH = icons_path

        # get and set app icon  
        app_icon_path = get_icon_path(app_icon_name) 
        if app_icon_path:
            self.setWindowIcon (QIcon (str(app_icon_path)))  

        # set dark or light mode
        scheme_name = Settings().get('color_scheme', Qt.ColorScheme.Unknown.name)   # either unknown (from System), Dark, Light
        QGuiApplication.styleHints().setColorScheme(Qt.ColorScheme[scheme_name])    # set scheme of QT

        # set mode for Widgets
        Widget.light_mode = not (scheme_name == Qt.ColorScheme.Dark.name)          


    def _set_win_geometry (self):
        """ set window geometry from settings """

        app_settings = Settings()                      # load app settings

        geometry = app_settings.get('window_geometry', [])
        maximize = app_settings.get('window_maximize', False)
        Win_Util.set_initialWindowSize (self, size_frac= (0.80, 0.80), pos_frac=(0.1, 0.1),
                                        geometry=geometry, maximize=maximize)


    def _save_app_settings (self):
        """ save application settings to file """

        s = Settings()

        s.set ('window_maximize', self.isMaximized())
        s.set ('window_geometry', self.normalGeometry().getRect())

        airfoil = self._app_model.airfoil
        if airfoil and not airfoil.isExample: 
            s.set ('last_opened', airfoil.pathFileName_abs)
        else:
            s.set ('last_opened', None)

        s.save()


    @override
    def closeEvent  (self, event : QCloseEvent):
        """ main window is closed """

        # save airfoil settings in app settings
        self._app_model.save_settings (to_app_settings=True,
                                       add_key  = self._diagram.name, 
                                       add_value= self._diagram.settings())

        # terminate polar watchdog thread, clean up working dir 
        self._app_model.close()                            # finish app model

        # save e.g. diagram options 
        self._save_app_settings () 

        event.accept()


#-------------------------------------------------------------------------------


def start ():
    """ start the app """

    # init logging - can be overwritten within a module  

    init_logging (level= logging.INFO)             # INFO, DEBUG or WARNING

    # command line arguments? 
    
    parser = argparse.ArgumentParser(prog=APP_NAME, description='View and modify an airfoil')
    parser.add_argument("airfoil", nargs='*', help="Airfoil .dat or .bez file to show")
    args = parser.parse_args()
    if args.airfoil: 
        initial_file = args.airfoil[0]
    else: 
        initial_file = None

    # init Qt Application and style  

    logger.info (f"Starting {APP_NAME} v{__version__} - Initial file: {initial_file}")

    app = QApplication(sys.argv)
    app.setStyle('fusion')

    main = Main (initial_file)
    main.show()
    rc = app.exec()
    return rc 



if __name__ == "__main__":
    
    sys.exit (start())